/*
    Legacy (trunk) - a portable emulator of the 8086-based IBM PC series of computers
    Copyright (C) 2005 Jonathan Thomas
    Contributions and fixes to CPU emulation by Jim Shaw

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include"cpu86.h"
#include"stdio.h"
#include"VideoDriverPalette13h.h"
#include"VideoDriverPalette1h.h"
#include"VideoDriverPalette3h.h"
#include"VideoDriverPaletteUndocumented.h"
#include"VideoDriverPalette4h.h"
#include"VideoDriverPalette5h.h"
#include"VideoDriverPalette6h.h"
#include"DiskImage.h"
#include"DebugOutput.h"
#include"MemoryMap.h"
#include"pic.h"

extern CVideoDriver *theVideo;
extern CDiskImage *theFloppy;
extern CDebugOutput *theDebugOutput;
extern CPic *thePic;
extern uint8 awaitingKeypress;
extern uint8 *asciimap;
extern uint8 port61h;
extern int exitkey;

//#define printf //printf

uint32 Cpu86::nativeInterrupt(uint8 intNumber) {

	uint16 destSeg;
	uint16 destOfs;
	uint16 index;
	uint8 *src;
	uint8 inkey;
	uint16 tailPointer;
	uint16 headPointer;
	uint16 tailPointerPlus2;

	static int interruptHLE=0;
	interruptHLE=0;

	//if (intNumber!=8) {
	//	printf("interrupt %x ah=%x al=%x\n",intNumber,ah,al);
	//}

	switch(intNumber) {
	case 0x3:
		//interruptHLE=1;

		break;
	case 0x8:
		setmemuint32(0,0x046c,getmemuint32(0,0x46c)+1);
		thePic->indicateNotBusy();
		break;
	case 0x1a:
		// get ticks since midnight
		dx=theCPU->memory[0x046c];
		cx=theCPU->memory[0x046e];
		alah.al=theCPU->memory[0x470];
		theCPU->memory[0x470]=0;
		//printf("ticks since midnight: %X\n",theCPU->getmemuint32(0,0x046c));
		interruptHLE=1;
		break;
	case 0x19:
		// soft reboot, don't erase interrupt vector table
		cs=0;
		ip=0x7c00;
		interruptHLE=1;
		break;
	case 0x20:
		interruptHLE=1;
		break;
	case 0x16:
		//printf("int 16h, subfunction %d\n",ah);
		switch(alah.ah) {
		case 0:
			// keyboard get keystroke

			headPointer=getmemuint16(0x40,0x1a);
			//if (headPointer!=getmemuint16(0x40,0x1c)) {

			// al contains ascii code, ah contains scan code

			alah.al=getmemuint8(0x40,getmemuint16(0x40,0x1a)+1);
			alah.ah=getmemuint8(0x40,getmemuint16(0x40,0x1a));

			headPointer+=2;	
			//printf("getting keystroke: %x\n",ah);

			if (headPointer>0x3c) {
				headPointer=0x1e;
			}
			setmemuint16(0x40,0x1a,headPointer);
			//} else {

				//printf("awaiting keypress\n");
				//awaitingKeypress=1;
				//while (getmemuint16(0x40,0x1a)==getmemuint16(0x40,0x1c)) {
					//printf("awaiting: %d\n",awaitingKeypress);
				//}

				//getchar();
				//printf("int16 ah=0 FIXME: wait for keyboard stroke\n");
				//printf("keyboard get keystroke, none available\n");
				// this should wait until a keypress is available.
				// wait for IRQ 1 to light up i guess
			//}

			interruptHLE=1;
			break;
		case 1:
			// keyboard check for keystroke
			// zeroFlag cleared if keystroke available

			zeroFlag=(getmemuint8(0x40,getmemuint16(0x40,0x1c))==getmemuint8(0x40,getmemuint16(0x40,0x1a)));
			setmemuint16(ss,sp+4,getmemuint16(ss,sp+4)&(65534<<6));
			setmemuint16(ss,sp+4,getmemuint16(ss,sp+4)|(zeroFlag<<6));

			if (zeroFlag) {
				ax=0;
			} else {
				//printf("int16 ah=1 keyboard check for keystroke: keystroke found\n");
				alah.al=getmemuint8(0x40,getmemuint16(0x40,0x1a)+1);
				alah.ah=getmemuint8(0x40,getmemuint16(0x40,0x1a));
			}

			//interruptHLE=1;

			//printf("keyboard check keystroke key %d, not available:%d, head %d, tail %d\n",ah,zeroFlag,getmemuint16(0x40,0x1a),getmemuint16(0x40,0x1c));

			break;
		case 2: // return shift flags in AL
			interruptHLE=1;
			alah.al=0;
			break;
		}
		break;
	case 0x9: // keyboard hit hardware interrupt

		// head pointer: 0040:001a
		// tail pointer: 0040:001c (where next keystroke is placed, incremented by 2 when key is entered
		// buffer: 0040:001e to 0040:003d

		// checkkbd:
		// mov ax,40h
		// mov ds,ax
		// mov ax,[1ah]
		// cmp ax,[1ch]
		// je checkkbd

		catchPortIn8(0x60,&inkey);
		if ((inkey>>7)==0) {
			tailPointer=getmemuint16(0x40,0x1c);
			//printf("inserting scancode into buffer scan[%x], ascii[%x], tail %d, head %d\n",inkey,asciimap[inkey],tailPointer,getmemuint16(0x40,0x1a));
			tailPointerPlus2=tailPointer+2;
			if (tailPointerPlus2>0x3c) {
				tailPointerPlus2=0x1e;
			}

			if (tailPointerPlus2!=getmemuint16(0x40,0x1a)) {
				setmemuint8(0x40,getmemuint16(0x40,0x1c),inkey);
				setmemuint8(0x40,getmemuint16(0x40,0x1c)+1,asciimap[inkey]);
				tailPointer+=2;	
				if (tailPointer>0x3c) {
					tailPointer=0x1e;
				}
				setmemuint16(0x40,0x1c,tailPointer);
			}
		}
		port61h&=127;
		thePic->indicateNotBusy();
		break;
	case 0x17: // printer services:
		alah.ah=192;
		interruptHLE=1;
		break;
	case 0x10: // video interrupt
		//printf("Video Interrupt ah=%x al=%x [%c,%c]\n",ah,al,ah,al);
		//sprintf(theDebugOutput->getBuffer(),"Interrupt 10, al=%x, ah=%x, bl=%x, bh=%x, cl=%x, ch=%x, dl=%x, dh=%x\n",*al,*ah,*bl,*bh,*cl,*ch,*dl,*dh);
		//theDebugOutput->writeMessage();
		switch(alah.ah) {
		case 0: // set video mode
			printf("set video mode: %x",alah.al);
			//getchar();
			switch(alah.al&127) {
			case 0x0:
			case 0x1:
					if (((alah.al)&128)==0) {
						memset(&memory[0xb8000],0,64000);
						printf("clearing video memory\n");
						for (index=0; index<64000; index+=2) {
							memory[0xb8001+index]=0x07;
						}
					}
					delete theVideo;
					theVideo=new CVideoDriverPalette1h(&theCPU->memory[0xb8000]);
					memory[MM_CURRENT_VIDEO_MODE_BYTE]=0x1;
					printf("Textmode 40x25\n");
					interruptHLE=1;
					// 40x25 text, 16 colours, 8 pages
				break;
			case 0x2:
			case 0x3:
					// 80x25 text, 16 colours, 8 pages
					if (((alah.al)&128)==0) {
						printf("clearing video memory\n");
						memset(&memory[0xb8000],0,64000);
						for (index=0; index<64000; index+=2) {
							memory[0xb8001+index]=0x07;
						}
					}
					delete theVideo;
					theVideo=new CVideoDriverPalette3h(&theCPU->memory[0xb8000]);
					memory[MM_CURRENT_VIDEO_MODE_BYTE]=0x3;
					interruptHLE=1;
					printf("Textmode 80x25\n");
				break;
			case 0x4:
					printf("al = %d, clear = %d\n",alah.al,((alah.al)&128)==0);
					if (((alah.al)&128)==0) {
						memset(&memory[0xb8000],0,64000);
						printf("clearing video memory\n");
					}
					delete theVideo;
					//theVideo=new CVideoDriverPalette4h(0xb8000);
					theVideo=new CVideoDriverPalette4h(&theCPU->memory[0xb8000]);
					memory[MM_CURRENT_VIDEO_MODE_BYTE]=0x4;
					interruptHLE=1;
					break;
			case 0x5:
					if (((alah.al)&128)==0) {
						memset(&memory[0xb8000],0,64000);
						printf("clearing video memory\n");
					}
					delete theVideo;
					theVideo=new CVideoDriverPalette4h(&theCPU->memory[0xb8000]);
					memory[MM_CURRENT_VIDEO_MODE_BYTE]=0x4;
					interruptHLE=1;
					break;
			case 0x6:
					if (((alah.al)&128)==0) {
						memset(&memory[0xb8000],0,64000);
						printf("clearing video memory\n");
					}
					delete theVideo;
					theVideo=new CVideoDriverPalette6h(&theCPU->memory[0xb8000]);
					memory[MM_CURRENT_VIDEO_MODE_BYTE]=0x6;
					interruptHLE=1;
					break;
			case 0x13:					// 320x200 Graphics, 256 colors, 1 page
					//printf("Switching to VGA 320x200x256\n");
					if (((alah.al)&128)==0) {
						memset(&memory[0xa0000],0,64000);
						printf("clearing video memory\n");
					}
					delete theVideo;
					theVideo=new CVideoDriverPalette13h(&theCPU->memory[0xA0000]);
					memory[MM_CURRENT_VIDEO_MODE_BYTE]=0x13;
					interruptHLE=1;
				break;
			case 0xd:
					// ega, 320x200, presumably 16 colours (a000)

				break;
			case 0xe:
					// ega, 640x200, not sure about colours (a000)
				break;
			case 0xf:
					// ega, 640x350, mono (a000)
				break;
			case 0x10:
					printf("Switching to EGA 640x350\n");
					interruptHLE=1;
					// ega, 640x350 a000 (16 colours, bitplaned)
					// this will need a memoryWatchZone to be set up i think				
					// http://webclass.cqu.edu.au/Units/81120_FOCT_Hardware/Study_Material/Study_Guide/chap2/chapter2.html
				break;
			case 0x12:
					printf("Switching to VGA 640x480x16\n");
					interruptHLE=1;
				break;
			}
			break;
		case 0x1: // define cursor appearance
			theVideo->setCursorSize(clch.ch, clch.cl);
			interruptHLE=1;
			break;
		case 0x2: // set cursor position:
			//printf("Set Cursor Position page %d column %d row %d\n",bh,dl,dh);
			theVideo->setCursorPosition(blbh.bh,dldh.dl,dldh.dh);
			interruptHLE=1;
			break;
		case 0x3:
			CURSORPOSITIONANDSIZE cursorPositionAndSize;
			theVideo->getCursorPosition(blbh.bh,&cursorPositionAndSize);
			dldh.dl=cursorPositionAndSize.column;
			dldh.dh=cursorPositionAndSize.row;
			clch.cl=cursorPositionAndSize.topScanLine;
			clch.ch=cursorPositionAndSize.bottomScanLine;
			interruptHLE=1;
			break;
		case 0x5: // set current display page:
			theVideo->setActivePage(alah.al);
			interruptHLE=1;
		case 0x6: // scroll text lines up
			theVideo->scrollActivePageUp(clch.ch,clch.cl,dldh.dh,dldh.dl,blbh.bh,alah.al);
			interruptHLE=1;
			break;
		case 0x7: // scroll text lines up
			theVideo->scrollActivePageDown(clch.ch,clch.cl,dldh.dh,dldh.dl,blbh.bh,alah.al);
			interruptHLE=1;
			break;
		case 0x8:
			CHARACTERATTRIBUTE characterAttribute;
			theVideo->readCharacterAndAttribute(blbh.bh,&characterAttribute);
			alah.al=characterAttribute.character;
			alah.ah=characterAttribute.attribute;
			interruptHLE=1;
			break;
		case 0x9:
			theVideo->writeCharacterAndAttribute(alah.al,blbh.bl,blbh.bh,cx);
			interruptHLE=1;
			break;
		case 0xa:
			theVideo->writeCharacterAtCursor(blbh.bh,alah.al,cx);
			interruptHLE=1;
			break;
		case 0xb:
			// colour palette
			interruptHLE=1;
			break;
		case 0xc:
			// sets the colour number of the specified pixel in graphics mode
			// guessing this just writes directly to vidmem, need to check
			interruptHLE=1;
			break;
		case 0xd:
			// reads the colour number of the specified pixel in graphics mode
			// guessing this just reads directly from vidmem, need to check
			interruptHLE=1;
			break;
		case 0xe:
			//printf("teletype write character (al=%d,bl=%d)\n",al,bl);
			theVideo->teletypeWriteCharacter(alah.al,blbh.bl);
			interruptHLE=1;
			break;
		case 0x10:
			//printf("Unknown interrupt in DUNES.COM (palette?)\n");

			index=0;
			theVideo->catchPortOut8(0x3c8,bx*3);
			while (cx>0) {
				theVideo->catchPortOut8(0x3c9,theCPU->getmemuint8(es,dx+index));
				cx--;
				index++;
			}

			interruptHLE=1;

			// vga colour functions, palette and border
			// definitely need to be highlevel emulated
			// or could be passed down to port level for "fun"
			break;
		/*case 0x11:
			interruptHLE=1;
			break;
		case 0x12:
			ax=getmemuint16(0x0040,0x0013);
			interruptHLE=1;
			break;*/
		case 0xf:
			// return current video parameters
			// this will need to be done properly
			alah.al=3;
			alah.ah=80;
			blbh.bh=0;
			interruptHLE=1;
			break;
			
		}
		break;
		case 0x11: // BIOS machine information
			ax=theCPU->getmemuint16(0x40,0x100);
			interruptHLE=1;
			break;
		case 0x12: // get memory size
			printf("get memory size");
			ax=getmemuint16(0x0040,0x0013);
			interruptHLE=1;
			break;
		case 0x21:
			printf("Called INT 21h, subfunction %x\n",alah.ah);
			exitkey=1;
			break;
		case 0x13:
			//printf("interrupt 0x13, function %d\n",ah);
			
			char *floppyBuffer;
			switch (alah.ah) {
			case 0:
				alah.ah=0;
				carryFlag=0;
				interruptHLE=1;
				//printf("reset disk system\n");
				//getchar();
				break;
			case 1:
				if (theFloppy!=NULL) {
					alah.ah=theFloppy->getError();
					carryFlag=theFloppy->getError()!=0;
					interruptHLE=1;
				}
				break;
			case 2:
				if (theFloppy!=NULL) {
					printf("drive number %d\n",dldh.dl);
					floppyBuffer=new char[alah.al*SECTOR_SIZE];
					alah.al=theFloppy->loadData(floppyBuffer,dldh.dh,clch.ch,clch.cl,alah.al);
					alah.ah=theFloppy->getError();
					//setmemuint16(*ss,(*sp)+4, (getmemuint16(*ss,(*sp)+4)&65534)|((*ah)!=0) );
					setmemuint16(ss,sp+4, (getmemuint16(ss,sp+4)&65534)|(alah.ah!=0) );
					sprintf(theDebugOutput->getBuffer(),"errorcode: %d\n",theFloppy->getError());
					theDebugOutput->writeMessage();

					//printf("sectors read: %d, error: %d\n",*al,*ah);
					printf("es=%x bx=%x\n",es,bx);
					//memcpy(getptrmemuint8(es,bx),floppyBuffer,al*SECTOR_SIZE);
					src=(uint8 *)floppyBuffer;
					destSeg=es;
					destOfs=bx;
					index=alah.al*SECTOR_SIZE;
					while (index>0) {
						setmemuint8(destSeg,destOfs,*src);
						destOfs++;
						index--;
						src++;
					}

					//printf("Bytes copied %d, first byte %X, %X\n",*al*SECTOR_SIZE,floppyBuffer[0],getmemuint8(*es,*bx));
					delete []floppyBuffer;
					interruptHLE=1;
					//getchar();

				}


				break;
			case 3:
				//*ah=theFloppy->saveData(getptrmemuint8(*es,*bx),*dh,*ch,*cl,*al);
				interruptHLE=1;
				break;
			case 4:
				interruptHLE=1;
				break;
			default:
				alah.ah=1;
			}
			break;


// end of fdd
	
	}


	return(interruptHLE);
}

